use std::{collections::HashMap, io, sync::Arc};
use tokio::{
    io::{AsyncBufReadExt, AsyncWriteExt, BufReader},
    net::{TcpListener, TcpStream},
    select, spawn,
    sync::{
        broadcast::{self, Receiver, Sender},
        Mutex,
    },
};
use uuid::Uuid;

#[derive(Clone, Debug)]
enum BroadcastMessage {
    MSG(String, String, String),
    NICK(String, String, String),
    JOIN(String),
    LEAVE(String),
    // we can add more types here, for example a graceful shutdown message
}

#[derive(Clone, Debug)]
enum ClientMessage {
    // Messages sent from the server to one client. This should not be broadcasted to other clients.
    ERR(String),
    LIST(Vec<String>),
}

#[tokio::main(flavor = "current_thread")]
async fn main() -> io::Result<()> {
    tracing_subscriber::fmt::init();

    let listener = TcpListener::bind("127.0.0.1:9090").await?;
    let (br_send, _) = broadcast::channel::<BroadcastMessage>(1024);
    let users: Arc<Mutex<HashMap<String, String>>> = Arc::new(Mutex::new(HashMap::new()));

    loop {
        let (socket, _) = listener.accept().await?;
        let (tx, rx) = (br_send.clone(), br_send.subscribe());
        let users = users.clone();

        spawn(async move {
            if let Err(e) = connection_handler(socket, tx, rx, users).await {
                tracing::error!("{:?}", e);
            }
        });
    }
}

async fn connection_handler(
    mut sock: TcpStream,
    br_sender: Sender<BroadcastMessage>,
    mut br_receiver: Receiver<BroadcastMessage>,
    users: Arc<Mutex<HashMap<String, String>>>,
) -> anyhow::Result<()> {
    tracing::info!("new client joined");

    let id = Uuid::new_v4().to_hyphenated().to_string();
    let mut nick = String::from(id.split("-").collect::<Vec<&str>>()[0]);

    tracing::info!("ID: {}, nick: {}", id, nick);

    users.lock().await.insert(id.clone(), nick.clone());
    br_sender.send(BroadcastMessage::JOIN(nick.clone()))?;

    let (rx, mut tx) = sock.split();
    let mut bufreader = BufReader::new(rx);

    let (cm_sender, mut cm_reciever) = broadcast::channel::<ClientMessage>(1024);

    loop {
        let mut buf = String::new();

        select! {
            o = bufreader.read_line(&mut buf) => {
                let bytes = o?;
                if bytes == 0 {
                    break;
                }

                buf = buf.trim().to_string();

                if buf == "" {
                    continue;
                }

                let (prefix, arguments) = {
                    let mut parts = buf.split_whitespace().map(|s| s.to_string()).collect::<Vec<String>>();

                    let prefix = parts.remove(0);

                    (prefix, parts)
                };

                tracing::info!("sender: {}, prefix: {}, arguments: {:#?}", nick, prefix, arguments.join(" "));

                match prefix.as_str() {
                    "MSG" => {
                        if arguments.len() < 1 {
                            cm_sender.send(ClientMessage::ERR(String::from("MSG requires at least 1 argument")))?;
                        } else {
                            br_sender.send(BroadcastMessage::MSG(nick.clone(), arguments.join(" "), id.clone()))?;
                        }
                    },
                    "NICK" => {
                        if let Some(new_nick) = arguments.iter().next() {
                            let old_nick = nick.clone();

                            let mut should_change = true;
                            for (_, nick) in users.lock().await.iter() {
                                if nick == new_nick {
                                    should_change = false;
                                    break;
                                }
                            }
                            if should_change {
                                nick = new_nick.to_owned();
                                users.lock().await.insert(id.clone(), nick.clone());
                                br_sender.send(BroadcastMessage::NICK(old_nick, nick.clone(), id.clone()))?;
                            } else {
                                cm_sender.send(ClientMessage::ERR(format!("Nickname {} is already taken!", new_nick)))?;
                            }
                        } else {
                            cm_sender.send(ClientMessage::ERR(String::from("NICK requires at least 1 argument")))?;
                        }
                    },
                    "LIST" => {
                        let mut list = Vec::new();
                        let current_nick = &nick;

                        for (_, nick) in users.lock().await.iter() {
                            if nick != current_nick {
                                list.push(nick.clone());
                            }
                        }
                        cm_sender.send(ClientMessage::LIST(list))?;
                    },
                    _ => {
                        cm_sender.send(ClientMessage::ERR(format!("Unknown command: {}", prefix)))?;
                    }
                }
            },
            o = br_receiver.recv() => {
                let msg = o?;
                match msg {
                    BroadcastMessage::MSG(nick, msg, sender) => {
                        if sender != id {
                            tx.write_all(format!("{} MSG {}\n", nick, msg).as_bytes()).await?;
                        }
                    },
                    BroadcastMessage::NICK(old_nick, new_nick, sender) => {
                        if sender != id {
                            tx.write_all(format!("{} NICK {}\n", old_nick, new_nick).as_bytes()).await?;
                        }
                    },
                    BroadcastMessage::JOIN(sender) => {
                        if sender != nick {
                            tx.write_all(format!("JOIN {}\n", sender).as_bytes()).await?;
                        }
                    },
                    BroadcastMessage::LEAVE(sender) => {
                        if sender != nick {
                            tx.write_all(format!("LEAVE {}\n", sender).as_bytes()).await?;
                        }
                    },
                }
            },
            o = cm_reciever.recv() => {
                let msg = o?;
                match msg {
                    ClientMessage::ERR(msg) => {
                        tracing::error!("ERR triggered by {}: {}", nick, msg);
                        tx.write_all(format!("ERR {}\n", msg).as_bytes()).await?;
                    },
                    ClientMessage::LIST(list) => {
                        let names = list.join(" ");

                        tx.write_all(format!("LIST {}\n", names).as_bytes()).await?;
                    }
                }
            }
        }
    }
    users.lock().await.remove(&id);
    br_sender.send(BroadcastMessage::LEAVE(nick))?;
    tracing::info!("Client {} disconnected", id);
    Ok(())
}
